
; Code to turn PROC lambdas into JUMP lambdas.

(define (integrate-jump-procs!)
  (receive (hits useless)
      (find-jump-procs (filter proc-lambda? (make-lambda-list)) find-calls)
    (remove-unused-procedures! useless)
    (for-each (lambda (p)
		(procs->jumps (cdr p)
			      (map bound-to-variable (cdr p))
			      (car p)))
	      hits)
    (not (and (null? hits) (null? useless)))))

; Make a call graph with extra nodes inserted for continuations:
;
; If F calls G tail-recursively, add an edge F->G
; If F calls G ... with continuation K, add a node K and edges F->K, K->G ...
;
; Then FIND-JOINS will return a list of the nodes that are passed two or
; more distinct continuations.  The rest can be merged with their callers.
;
; Need a root node, so make one that points to all procs with unknown calls.

(define-record-type node
  (proc             ; lambda node (or #f for continuation holders)
   cont)            ; lambda node (or #f for procs)
  ((successors '())
   temp
   (join? #f)
   (merged? #f)
   ))

(define-record-discloser type/node
  (lambda (node)
    (list 'node (node-proc node) (node-cont node))))

(define (add-child! parent child)
  (if (not (memq? child (node-successors parent)))
      (set-node-successors! parent (cons child (node-successors parent)))))

(define (find-jump-procs all-procs proc->uses)
  (for-each (lambda (l)
	      (set-lambda-block! l (node-maker l #f)))
	    all-procs)
  (receive (known unknown)
      (partition-list calls-known? all-procs)
    (let ((root (node-maker #f #f))
	  (conts-cell (list '()))
	  (known-blocks (map lambda-block known))
	  (procs-cell (list (map lambda-block unknown))))
      (note-calls! known conts-cell procs-cell proc->uses)
      (let ((unknown-blocks (car procs-cell)))
	(set-node-successors! root unknown-blocks)
	(graph->ssa-graph! root node-successors node-temp set-node-temp!)
	(let ((joins (find-joins (car conts-cell) node-temp)))
	  (for-each (lambda (n)
		      (set-node-join?! n #t))
		    joins)
	  (let* ((mergable (filter-map find-mergable
				       (append joins
					       unknown-blocks
					       (car conts-cell))))
		 (useless  (filter (lambda (p)
				     (not (or (node-join? (lambda-block p))
					      (node-merged? (lambda-block p)))))
				   known)))
	    (for-each (lambda (p)
			(set-lambda-block! p #f))
		      all-procs)
	    (values mergable useless)))))))

(define (note-calls! known-procs conts-cell procs-cell proc->uses)
  (for-each (lambda (proc)
	      (for-each (lambda (ref)
			  (note-call! (lambda-block proc)
				      ref
				      conts-cell procs-cell))
			(proc->uses proc)))
	    known-procs))

(define (note-call! proc-node ref conts-cell procs-cell)
  (let ((caller (get-lambda-block (containing-procedure ref) procs-cell)))
    (if (calls-this-primop? (node-parent ref) 'tail-call)
	(add-child! caller proc-node)
	(note-non-tail-call! proc-node
			     caller
			     (call-arg (node-parent ref) 0)
			     conts-cell))))

(define (get-lambda-block proc procs-cell)
  (let ((block (lambda-block proc)))
    (if (node? block)
	block
	(let ((new (node-maker proc #f)))
	  (set-lambda-block! proc new)
	  (set-car! procs-cell (cons new (car procs-cell)))
	  new))))

(define (note-non-tail-call! proc-node caller cont conts-cell)
  (let loop ((succs (node-successors caller)))
    (if (null? succs)
	(let ((new (node-maker #f cont)))
	  (set-car! conts-cell (cons new (car conts-cell)))
	  (set-node-successors! new (list proc-node))
	  (set-node-successors! caller (cons new (node-successors caller))))
	(let ((node (car succs)))
	  (if (and (node-cont node)
		   (node-equal? cont (node-cont node)))
	      (add-child! node proc-node)
	      (loop (cdr succs)))))))

(define (find-mergable node)
  (let ((mergable (really-find-mergable node)))
    (if (null? mergable)
	#f
	(cons (or (node-cont node)
		  (car (variable-refs (car (lambda-variables (node-proc node))))))
	      mergable))))

(define (really-find-mergable node)    
  (let recur ((nodes (node-successors node)) (res '()))
    (if (null? nodes)
	res
	(recur (cdr nodes)
	       (let ((node (car nodes)))
		 (cond ((or (node-join? node)     ; gets two or more continuations
			    (node-merged? node)   ; already merged
			    (node-cont node))     ; different continuation
			res)
;		       ((node-cont node)          ; not a lambda
;			(recur (node-successors node) res))
		       (else
			(set-node-merged?! node #t)
			(recur (node-successors node)
			       (cons (node-proc node) res)))))))))

;=============================================================================;
; Part 2.  PROCS is a list of procedures that are only called by each other;
; with no entry point they are useless and can be removed.

(define (remove-unused-procedures! procs)
  (for-each (lambda (proc)
	      (let ((var (bound-to-variable proc)))
		(if (not var)
		    (bug "known procedure has no variable ~S" proc))
		(format #t "Removing unused procedure: ~S~%"
			(variable-name var)) ; would LAMBDA-NAME be better?
		(mark-changed (node-parent proc))
		(detach-bound-value var proc)
		(erase proc)))
	    procs))

;=============================================================================;
; Part 3.  Turn JUMP-PROCS from procs to jumps.  CONT is the continuation they
; all receive, and is also turned into a jump.

; This finds the least-common-ancestor of all of the calls to the
; jump-procs and binds them and the continuation (if any) with a LETREC at
; that point.  VARS are the variables to which JUMP-PROCS are bound.  If
; there is a continuation some protocol adjustment may be required.

(define (procs->jumps jump-procs vars cont)
;  (if (null? (filter (lambda (n)
;                         (node-equal? cont (call-arg (node-parent n) 0)))
;                       (apply append (map variable-refs vars))))
;      (breakpoint "no calls"))
  (let* ((proc (containing-procedure cont))
	 (lca (least-common-ancestor
	       (filter (lambda (n)
			 (node-equal? cont (call-arg (node-parent n) 0)))
		       (apply append (map variable-refs vars)))))
	 (lca (if (call-node? lca) lca (node-parent lca)))
	 (cvar (if (lambda-node? cont)
		   (make-variable 'w (node-type cont))
		   #f))
	 (proc-cvar (car (lambda-variables proc))))
    (receive (vars jump-procs)
	(bound-above? lca vars jump-procs)
      (for-each detach-bound-value vars jump-procs)
      (cond ((lambda-node? cont)
	     (determine-continuation-protocol cont jump-procs)
	     (move cont (lambda (ignore) (make-literal-node '#f '#f)))
	     (put-in-letrec (cons cvar vars) (cons cont jump-procs) lca)
	     (change-lambda-type cont 'jump))
	    (else
	     (put-in-letrec vars jump-procs lca))))
    (for-each proc-calls->jumps jump-procs)
    (for-each (lambda (p)
		(let* ((v (car (lambda-variables p)))
		       (refs (variable-refs v)))
		  (set-variable-refs! v '())
		  (for-each (lambda (r)
			      (if (lambda-node? cont)
				  (return->jump (node-parent r) cvar cont)
				  (replace r (make-reference-node proc-cvar))))
			    refs)
		  (remove-variable p v)))
	      jump-procs)
    (values)))

; Return the list of VARS and VALS where the variable is either global
; or bound above CALL.

(define (bound-above? call vars vals)
  (set-node-flag! call #t)
  (let loop ((vars vars) (vals vals) (r-vars '()) (r-vals '()))
    (cond ((null? vars)
	   (set-node-flag! call #f)
	   (values r-vars r-vals))
	  ((and (variable-binder (car vars))
		(marked-ancestor (variable-binder (car vars))))
	   (loop (cdr vars) (cdr vals) r-vars r-vals))
	  (else
	   (loop (cdr vars) (cdr vals)
		 (cons (car vars) r-vars)
		 (cons (car vals) r-vals))))))

(define (detach-bound-value var node)
  (if (variable-binder var)
      (let ((binder (variable-binder var))
	    (parent (node-parent node))
	    (index (node-index node)))
	(set-lambda-variables! binder (delq! var (lambda-variables binder)))
	(detach node)
	(remove-call-arg parent index))))
  
; Turn all calls to PROC into jumps.

(define (proc-calls->jumps proc)
  (for-each (lambda (n)
	      (call->jump (node-parent n)))
	    (find-calls proc))
  (change-lambda-type proc 'jump))

; Change a call to a jump by changing the primop and removing the continuation.

(define (call->jump call)
  (case (primop-id (call-primop call))
    ((call tail-call)
     (set-call-primop! call (get-primop (enum primop jump)))
     (remove-call-arg call 0))
    (else
     (bug "odd call primop ~S" (call-primop call)))))

; Change a return to a jump.  VAR is a variable bound to JUMP, the lambda
; being jumped to.

(define (return->jump call var jump)
  (case (primop-id (call-primop call))
    ((return)
     (set-call-primop! call (get-primop (enum primop jump)))
     (replace (call-arg call 0) (make-reference-node var)))
    (else
     (bug "odd return primop ~S" (call-primop call)))))
